@@ -9,14 +9,19 @@ import plotly
 from pkg_resources import parse_version
 
 import dtale.dash_application.custom_geojson as custom_geojson
+import dtale.dash_application.extended_aggregations as extended_aggregations
 import dtale.global_state as global_state
 from dtale.charts.utils import (
+    AGGS,
     ANIMATION_CHARTS,
     ANIMATE_BY_CHARTS,
     YAXIS_CHARTS,
     ZAXIS_CHARTS,
+    NON_EXT_AGGREGATION,
+    build_final_cols,
     find_group_vals,
 )
+from dtale.column_builders import get_cleaner_configs
 from dtale.dash_application.layout.utils import (
     build_input,
     build_option,
@@ -25,7 +30,6 @@ from dtale.dash_application.layout.utils import (
     build_hoverable,
     build_selections,
     show_style,
-    AGGS,
     FREQS,
     FREQ_LABELS,
 )
@@ -178,6 +182,7 @@ CHARTS = [
     dict(value="maps"),
     dict(value="candlestick"),
     dict(value="treemap"),
+    dict(value="funnel"),
 ]
 CHART_INPUT_SETTINGS = {
     "line": dict(
@@ -252,8 +257,18 @@ CHART_INPUT_SETTINGS = {
         cs_group=dict(display=False),
         treemap_group=dict(display=True),
     ),
+    "funnel": dict(
+        x=dict(display=False),
+        y=dict(display=False),
+        z=dict(display=False),
+        group=dict(display=False),
+        map_group=dict(display=False),
+        cs_group=dict(display=False),
+        treemap_group=dict(display=False),
+        funnel_group=dict(display=True),
+    ),
 }
-
+LOAD_TYPES = ["random", "head", "tail"]
 MAP_TYPES = [
     dict(value="choropleth", image=True),
     dict(value="scattergeo", label="ScatterGeo", image=True),
@@ -290,11 +305,51 @@ def auto_load_msg():
     return text("auto_load_msg")
 
 
+def load_type_msg():
+    return [
+        html.Div(
+            [
+                html.H3(
+                    text("Load Types"), style=dict(display="inline"), className="pr-3"
+                ),
+            ]
+        ),
+        html.Ul(
+            [
+                html.Li(
+                    [
+                        html.B(text("Random")),
+                        html.Span(": {}".format(text("load_random"))),
+                    ],
+                    className="mb-0",
+                ),
+                html.Li(
+                    [
+                        html.B(text("Head")),
+                        html.Span(": {}".format(text("load_head"))),
+                    ],
+                    className="mb-0",
+                ),
+                html.Li(
+                    [
+                        html.B(text("Tail")),
+                        html.Span(": {}".format(text("load_tail"))),
+                    ],
+                    className="mb-0",
+                ),
+            ],
+            className="mb-0",
+        ),
+    ]
+
+
 def bin_type_msg():
     return [
         html.Div(
             [
-                html.H3("Bin Types", style=dict(display="inline"), className="pr-3"),
+                html.H3(
+                    text("Bin Types"), style=dict(display="inline"), className="pr-3"
+                ),
                 html.A(
                     "({})".format(text("Binning in Data Mining")),
                     href="https://www.geeksforgeeks.org/binning-in-data-mining/",
@@ -554,24 +609,37 @@ def get_group_types(inputs, group_cols=None):
         return _flags("cs_group")
     elif show_input_handler(chart_type)("treemap_group"):
         return _flags("treemap_group")
+    elif show_input_handler(chart_type)("funnel_group"):
+        return _flags("funnel_group")
     return False, False
 
 
-def update_label_for_freq(val):
+def update_label_for_freq_and_agg(val):
     """
-    Formats sub-values contained within 'val' to display date frequencies if included.
+    Formats sub-values contained within 'val' to display date frequencies & aggregatioms if included.
         - (val=['a', 'b', 'c']) => 'a, b, c'
-        - (val=['a|H', 'b', 'c']) => 'a (Hour), b, c'
+        - (val=['a|H', 'b|mean', 'c']) => 'a (Hour), Mean of b, c'
     """
 
     def _freq_handler(sub_val):
+        selected_agg = None
+        for agg in AGGS:
+            if sub_val.endswith("|{}".format(agg)):
+                selected_agg = "{} of".format(text(AGGS[agg]))
+                sub_val = sub_val.split("|{}".format(agg))[0]
+                break
+
+        selected_freq = None
         for freq in FREQS:
             if sub_val.endswith("|{}".format(freq)):
                 col, freq = sub_val.split("|")
-                if freq in FREQ_LABELS:
-                    return "{} ({})".format(col, text(FREQ_LABELS[freq]))
-                return col
-        return sub_val
+                if freq in FREQS:
+                    sub_val = sub_val.split("|{}".format(freq))[0]
+                    if freq in FREQ_LABELS:
+                        selected_freq = "({})".format(text(FREQ_LABELS[freq]))
+                    break
+
+        return " ".join(list(filter(None, [selected_agg, sub_val, selected_freq])))
 
     return ", ".join([_freq_handler(sub_val) for sub_val in make_list(val)])
 
@@ -601,7 +669,7 @@ def build_error(error, tb):
     )
 
 
-def build_input_options(df, **inputs):
+def build_input_options(df, extended_aggregation=[], **inputs):
     """
     Builds dropdown options for (X, Y, Z, Group, Barsort & Y-Axis Ranges) with filtering based on currently selected
     values for the following inputs: x, y, z, group.
@@ -629,9 +697,14 @@ def build_input_options(df, **inputs):
         for c, l in col_opts
         if c not in build_selections(x, y, z_val)
     ]
-    barsort_options = [build_option(o) for o in build_selections(x, y)]
-    yaxis_options = [build_option(y2) for y2 in y or []]
-
+    final_cols = build_final_cols(
+        y,
+        z,
+        inputs.get("agg"),
+        extended_aggregation if chart_type not in NON_EXT_AGGREGATION else [],
+    )
+    barsort_options = [build_option(o) for o in build_selections(x, final_cols)]
+    yaxis_options = [build_option(y2) for y2 in final_cols or []]
     return (
         x_options,
         y_multi_options,
@@ -722,7 +795,7 @@ def build_candlestick_options(
     return x_options, close_options, open_options, low_options, high_options
 
 
-def build_treemap_options(df, treemap_value=None, treemap_label=None):
+def build_label_value_options(df, selected_value=None, selected_label=None):
     dtypes = get_dtypes(df)
     cols = sorted(dtypes.keys())
     num_cols = []
@@ -733,14 +806,106 @@ def build_treemap_options(df, treemap_value=None, treemap_label=None):
             num_cols.append(c)
 
     value_options = [
-        build_option(c) for c in num_cols if c not in build_selections(treemap_label)
+        build_option(c) for c in num_cols if c not in build_selections(selected_label)
     ]
     label_options = [
-        build_option(c) for c in cols if c not in build_selections(treemap_value)
+        build_option(c) for c in cols if c not in build_selections(selected_value)
     ]
     return value_options, label_options
 
 
+def build_label_value_store(prop, inputs):
+    return dcc.Store(
+        id="{}-input-data".format(prop),
+        data={
+            k: v
+            for k, v in inputs.items()
+            if k
+            in [
+                "{}_value".format(prop),
+                "{}_label".format(prop),
+                "{}_group".format(prop),
+            ]
+        },
+    )
+
+
+def build_label_value_inputs(prop, inputs, df, group_options, additional_inputs=[]):
+    show_inputs = inputs.get("chart_type") == "treemap"
+    props = ["{}_value", "{}_label", "{}_group"]
+    selected_value, selected_label, selected_group = (
+        inputs.get(p.format(prop)) for p in props
+    )
+    (value_options, label_options,) = build_label_value_options(
+        df,
+        selected_value=selected_value,
+        selected_label=selected_label,
+    )
+    return html.Div(
+        [
+            build_input(
+                text("Value"),
+                dcc.Dropdown(
+                    id="{}-value-dropdown".format(prop),
+                    options=value_options,
+                    placeholder=text("Select Column"),
+                    style=dict(width="inherit"),
+                    value=selected_value,
+                ),
+            ),
+            build_input(
+                text("Labels"),
+                dcc.Dropdown(
+                    id="{}-label-dropdown".format(prop),
+                    options=label_options,
+                    placeholder=text("Select Column"),
+                    style=dict(width="inherit"),
+                    value=selected_label,
+                ),
+            ),
+        ]
+        + additional_inputs
+        + [
+            build_input(
+                text("Group"),
+                dcc.Dropdown(
+                    id="{}-group-dropdown".format(prop),
+                    options=group_options,
+                    multi=True,
+                    placeholder=text("Select Group(s)"),
+                    value=selected_group,
+                    style=dict(width="inherit"),
+                ),
+                className="col",
+                id="{}-group-input".format(prop),
+            ),
+        ],
+        id="{}-inputs".format(prop),
+        className="row charts-filters",
+        style={} if show_inputs else {"display": "none"},
+    )
+
+
+def build_funnel_inputs(inputs, df, group_options):
+    stacked_toggle = build_input(
+        "{}?".format(text("Stack")),
+        html.Div(
+            daq.BooleanSwitch(
+                id="funnel-stack-toggle",
+                on=False,
+            ),
+            className="toggle-wrapper",
+        ),
+        id="funnel-stack-input",
+        style=show_style(inputs.get("funnel_group") is not None),
+        className="col-auto",
+    )
+
+    return build_label_value_inputs(
+        "funnel", inputs, df, group_options, additional_inputs=[stacked_toggle]
+    )
+
+
 def build_mapbox_style_options():
     from dtale.charts.utils import get_mapbox_token
 
@@ -817,7 +982,7 @@ def show_chart_per_group(**inputs):
     Boolean function to determine whether "Chart Per Group" toggle should be displayed or not
     """
     chart_type, group = (inputs.get(p) for p in ["chart_type", "group"])
-    invalid_type = chart_type in ["pie", "wordcloud", "maps"]
+    invalid_type = chart_type in ["pie", "wordcloud", "maps", "funnel"]
     return (
         show_input_handler(chart_type)("group")
         and len(group or [])
@@ -944,8 +1109,8 @@ def build_slider_counts(df, data_id, query_value):
     return slider_counts
 
 
-def data_selection_text(is_open):
-    return "{} {}".format("\u25BC" if is_open else "\u25B6", text("Data Selection"))
+def collapse_btn_text(is_open, label):
+    return "{} {}".format("\u25BC" if is_open else "\u25B6", label)
 
 
 def charts_layout(df, settings, **inputs):
@@ -960,8 +1125,9 @@ def charts_layout(df, settings, **inputs):
     :type param: dict
     :return: dash markup
     """
-    chart_type, x, y, z, group, agg, load = (
-        inputs.get(p) for p in ["chart_type", "x", "y", "z", "group", "agg", "load"]
+    chart_type, x, y, z, group, agg, load, load_type = (
+        inputs.get(p)
+        for p in ["chart_type", "x", "y", "z", "group", "agg", "load", "load_type"]
     )
     loc_modes = loc_mode_info()
     y = y or []
@@ -1033,14 +1199,6 @@ def charts_layout(df, settings, **inputs):
         cs_high=cs_high,
         cs_low=cs_low,
     )
-    show_treemap = chart_type == "treemap"
-    treemap_props = ["treemap_value", "treemap_label", "treemap_group"]
-    treemap_value, treemap_label, treemap_group = (inputs.get(p) for p in treemap_props)
-    (treemap_value_options, treemap_label_options,) = build_treemap_options(
-        df,
-        treemap_value=treemap_value,
-        treemap_label=treemap_label,
-    )
     cscale_style = colorscale_input_style(**inputs)
     default_cscale = DEFAULT_CSALES.get(chart_type, REDS)
 
@@ -1100,14 +1258,8 @@ def charts_layout(df, settings, **inputs):
                 if k in ["cs_x", "cs_open", "cs_close", "cs_high", "cs_low", "cs_group"]
             },
         ),
-        dcc.Store(
-            id="treemap-input-data",
-            data={
-                k: v
-                for k, v in inputs.items()
-                if k in ["treemap_value", "treemap_label", "treemap_group"]
-            },
-        ),
+        build_label_value_store("treemap", inputs),
+        build_label_value_store("funnel", inputs),
         dcc.Store(id="range-data"),
         dcc.Store(id="yaxis-data", data=inputs.get("yaxis")),
         dcc.Store(id="last-chart-input-data", data=inputs),
@@ -1117,22 +1269,38 @@ def charts_layout(df, settings, **inputs):
         html.Div(
             [
                 html.Div(
-                    dbc.Button(data_selection_text(False), id="collapse-data-btn"),
+                    dbc.Button(
+                        collapse_btn_text(False, text("Data Selection")),
+                        id="collapse-data-btn",
+                    ),
                     className="col-auto pr-0",
                 ),
                 html.Div(
                     dbc.Collapse(
                         dbc.Card(
                             dbc.CardBody(
-                                dcc.Tabs(
-                                    id="data-tabs",
-                                    value=inputs["data_id"],
-                                    children=[
-                                        build_tab(global_state.get_name(k) or k, str(k))
-                                        for k in global_state.keys()
-                                    ],
-                                    style=dict(height="36px"),
-                                ),
+                                [
+                                    dcc.Tabs(
+                                        id="data-tabs",
+                                        value=inputs["data_id"],
+                                        children=[
+                                            build_tab(
+                                                global_state.get_name(k) or k,
+                                                str(k),
+                                                id="data-tab-{}".format(k),
+                                            )
+                                            for k in global_state.keys()
+                                        ],
+                                        style=dict(height="36px"),
+                                    )
+                                ]
+                                + [
+                                    dbc.Tooltip(
+                                        global_state.get_name(k) or k,
+                                        target="data-tab-{}".format(k),
+                                    )
+                                    for k in global_state.keys()
+                                ]
                             )
                         ),
                         id="collapse-data",
@@ -1145,17 +1313,28 @@ def charts_layout(df, settings, **inputs):
         ),
         html.Div(
             html.Div(
-                dcc.Tabs(
-                    id="chart-tabs",
-                    value=chart_type or "line",
-                    children=[
-                        build_tab(
-                            text(t.get("label", t["value"].capitalize())), t["value"]
-                        )
-                        for t in CHARTS
-                    ],
-                    style=dict(height="36px"),
-                ),
+                [
+                    dcc.Tabs(
+                        id="chart-tabs",
+                        value=chart_type or "line",
+                        children=[
+                            build_tab(
+                                text(t.get("label", t["value"].capitalize())),
+                                t["value"],
+                                id="chart-type-tab-{}".format(i),
+                            )
+                            for i, t in enumerate(CHARTS)
+                        ],
+                        style=dict(height="36px"),
+                    ),
+                ]
+                + [
+                    dbc.Tooltip(
+                        text(t.get("label", t["value"].capitalize())),
+                        target="chart-type-tab-{}".format(i),
+                    )
+                    for i, t in enumerate(CHARTS)
+                ],
                 className="col-md-12",
             ),
             className="row pt-3 pb-3 charts-filters",
@@ -1189,11 +1368,18 @@ def charts_layout(df, settings, **inputs):
                         [
                             build_hoverable(
                                 html.Span(text("Load")),
-                                html.Span(
-                                    text(
-                                        "Load a random percentage of rows from your dataset."
-                                    )
-                                ),
+                                load_type_msg(),
+                                additional_classes="load-types-tooltip mb-auto mt-auto",
+                                top="100%",
+                            ),
+                            dcc.Dropdown(
+                                id="load-type-dropdown",
+                                options=[
+                                    build_option(v, v.capitalize()) for v in LOAD_TYPES
+                                ],
+                                className="pl-5",
+                                value=load_type or "random",
+                                clearable=False,
                             ),
                             dcc.Slider(
                                 id="load-input",
@@ -1216,7 +1402,49 @@ def charts_layout(df, settings, **inputs):
                 ],
                 className="col",
             ),
-            className="row pt-3 pb-3 charts-filters",
+            className="row pt-3 pb-0 charts-filters",
+        ),
+        html.Div(
+            [
+                html.Div(
+                    [
+                        dbc.Button(
+                            collapse_btn_text(False, text("Cleaners")),
+                            id="collapse-cleaners-btn",
+                        ),
+                        html.Span(
+                            "",
+                            id="selected-cleaners",
+                            className="pl-3",
+                            style=dict(fontSize="85%"),
+                        ),
+                    ],
+                    className="col-auto pr-0",
+                ),
+                html.Div(
+                    dbc.Collapse(
+                        dbc.Card(
+                            dbc.CardBody(
+                                dcc.Dropdown(
+                                    id="cleaners-dropdown",
+                                    options=[
+                                        build_option(c["value"], c["label"])
+                                        for c in get_cleaner_configs()
+                                    ],
+                                    multi=True,
+                                    placeholder=text("Select Cleaner(s)"),
+                                    value=group,
+                                    style=dict(width="inherit"),
+                                ),
+                            )
+                        ),
+                        id="collapse-cleaners",
+                    ),
+                    className="col",
+                ),
+            ],
+            className="row pb-3 charts-filters",
+            style=dict(display="block"),
         ),
         html.Div(
             [
@@ -1305,7 +1533,7 @@ def charts_layout(df, settings, **inputs):
                             style={}
                             if not show_map
                             and not show_candlestick
-                            and not show_treemap
+                            and chart_type not in ["treemap", "funnel"]
                             else {"display": "none"},
                             className="row p-0 charts-filters",
                         ),
@@ -1546,80 +1774,74 @@ def charts_layout(df, settings, **inputs):
                             className="row charts-filters",
                             style={} if show_candlestick else {"display": "none"},
                         ),
+                        build_label_value_inputs("treemap", inputs, df, group_options),
+                        build_funnel_inputs(inputs, df, group_options),
                         html.Div(
                             [
-                                build_input(
-                                    text("Value"),
-                                    dcc.Dropdown(
-                                        id="treemap-value-dropdown",
-                                        options=treemap_value_options,
-                                        placeholder="Select a column",
-                                        style=dict(width="inherit"),
-                                        value=treemap_value,
-                                    ),
-                                ),
-                                build_input(
-                                    text("Labels"),
-                                    dcc.Dropdown(
-                                        id="treemap-label-dropdown",
-                                        options=treemap_label_options,
-                                        placeholder=text("Select Column"),
-                                        style=dict(width="inherit"),
-                                        value=treemap_label,
-                                    ),
-                                ),
-                                build_input(
-                                    text("Group"),
-                                    dcc.Dropdown(
-                                        id="treemap-group-dropdown",
-                                        options=group_options,
-                                        multi=True,
-                                        placeholder=text("Select Group(s)"),
-                                        value=inputs.get("treemap_group"),
-                                        style=dict(width="inherit"),
-                                    ),
-                                    className="col",
-                                    id="treemap-group-input",
-                                ),
-                            ],
-                            id="treemap-inputs",
-                            className="row charts-filters",
-                            style={} if show_treemap else {"display": "none"},
-                        ),
-                        html.Div(
-                            [
-                                build_input(
-                                    text("Aggregation"),
-                                    dcc.Dropdown(
-                                        id="agg-dropdown",
-                                        options=[
-                                            build_option(v, text(AGGS[v]))
-                                            for v in [
-                                                "count",
-                                                "nunique",
-                                                "sum",
-                                                "mean",
-                                                "rolling",
-                                                "corr",
-                                                "first",
-                                                "last",
-                                                "drop_duplicates",
-                                                "median",
-                                                "min",
-                                                "max",
-                                                "std",
-                                                "var",
-                                                "mad",
-                                                "prod",
-                                                "pctsum",
-                                                "pctct",
-                                            ]
-                                        ],
-                                        placeholder=text("No Aggregation"),
-                                        style=dict(width="inherit"),
-                                        value=agg or "raw",
-                                    ),
+                                html.Div(
+                                    [
+                                        build_input(
+                                            text("Aggregation"),
+                                            dcc.Dropdown(
+                                                id="agg-dropdown",
+                                                options=[
+                                                    build_option(v, text(AGGS[v]))
+                                                    for v in [
+                                                        "count",
+                                                        "nunique",
+                                                        "sum",
+                                                        "mean",
+                                                        "rolling",
+                                                        "corr",
+                                                        "first",
+                                                        "last",
+                                                        "drop_duplicates",
+                                                        "median",
+                                                        "min",
+                                                        "max",
+                                                        "std",
+                                                        "var",
+                                                        "mad",
+                                                        "prod",
+                                                        "pctsum",
+                                                        "pctct",
+                                                    ]
+                                                ],
+                                                placeholder=text("No Aggregation"),
+                                                style=dict(width="inherit"),
+                                                value=agg or "raw",
+                                                disabled=len(
+                                                    inputs.get(
+                                                        "extended_aggregation", []
+                                                    )
+                                                )
+                                                > 0,
+                                            ),
+                                        ),
+                                        html.Span(
+                                            "* {}".format(
+                                                text(
+                                                    "Extended Aggregation is currently populated."
+                                                )
+                                            ),
+                                            id="ext-agg-warning",
+                                            style=show_style(
+                                                len(
+                                                    inputs.get(
+                                                        "extended_aggreation", []
+                                                    )
+                                                )
+                                                > 0
+                                            ),
+                                            className="ext-agg-warning",
+                                        ),
+                                    ]
                                 ),
+                            ]
+                            + extended_aggregations.build_modal(
+                                inputs.get("extended_aggregation", []), chart_type, y
+                            )
+                            + [
                                 html.Div(
                                     [
                                         build_input(
