@@ -9,14 +9,20 @@ from dash.exceptions import PreventUpdate
 
 import dtale.dash_application.custom_geojson as custom_geojson
 import dtale.dash_application.drilldown_modal as drilldown_modal
+import dtale.dash_application.extended_aggregations as extended_aggregations
 import dtale.dash_application.saved_charts as saved_charts
 import dtale.dash_application.lock_zoom as lock_zoom
 import dtale.global_state as global_state
-from dtale.charts.utils import MAX_GROUPS, ZAXIS_CHARTS
+from dtale.charts.utils import (
+    MAX_GROUPS,
+    ZAXIS_CHARTS,
+    NON_EXT_AGGREGATION,
+    build_final_cols,
+)
+from dtale.code_export import CHART_EXPORT_CODE
 from dtale.dash_application.charts import (
     build_chart,
     chart_url_params,
-    CHART_EXPORT_CODE,
     valid_chart,
 )
 from dtale.dash_application.layout.layout import (
@@ -31,10 +37,10 @@ from dtale.dash_application.layout.layout import (
     build_mapbox_style_options,
     build_proj_hover_children,
     build_slider_counts,
-    build_treemap_options,
+    build_label_value_options,
     charts_layout,
     colorscale_input_style,
-    data_selection_text,
+    collapse_btn_text,
     get_yaxis_type_tabs,
     lock_zoom_style,
     main_inputs_and_group_val_display,
@@ -46,6 +52,7 @@ from dtale.dash_application.layout.layout import (
 )
 from dtale.dash_application.layout.utils import show_style
 from dtale.dash_application.utils import get_data_id
+from dtale.translations import text
 from dtale.query import run_query
 from dtale.utils import dict_merge, is_app_root_defined, make_list
 
@@ -197,12 +204,16 @@ def init_callbacks(dash_app):
             Output("map-inputs", "style"),
             Output("candlestick-inputs", "style"),
             Output("treemap-inputs", "style"),
+            Output("funnel-inputs", "style"),
             Output("colorscale-input", "style"),
             Output("drilldown-input", "style"),
             Output("lock-zoom-btn", "style"),
+            Output("open-extended-agg-modal", "style"),
+            Output("selected-cleaners", "children"),
         ],
         [
             Input("query-data", "modified_timestamp"),
+            Input("extended-aggregations", "modified_timestamp"),
             Input("chart-tabs", "value"),
             Input("x-dropdown", "value"),
             Input("y-multi-dropdown", "value"),
@@ -217,15 +228,19 @@ def init_callbacks(dash_app):
             Input("window-input", "value"),
             Input("rolling-comp-dropdown", "value"),
             Input("load-input", "value"),
+            Input("load-type-dropdown", "value"),
+            Input("cleaners-dropdown", "value"),
         ],
         [
             State("url", "pathname"),
             State("query-data", "data"),
             State("data-tabs", "value"),
+            State("extended-aggregations", "data"),
         ],
     )
     def input_data(
         _ts,
+        _ts2,
         chart_type,
         x,
         y_multi,
@@ -240,9 +255,12 @@ def init_callbacks(dash_app):
         window,
         rolling_comp,
         load,
+        load_type,
+        cleaners,
         pathname,
         query,
         data_id,
+        extended_aggregation,
     ):
         """
         dash callback for maintaining chart input state and column-based dropdown options.  This will guard against
@@ -269,8 +287,14 @@ def init_callbacks(dash_app):
             window=window,
             rolling_comp=rolling_comp,
             load=load,
+            load_type=load_type,
+            cleaners=make_list(cleaners),
+        )
+        options = build_input_options(
+            global_state.get_data(data_id),
+            extended_aggregation=extended_aggregation,
+            **inputs
         )
-        options = build_input_options(global_state.get_data(data_id), **inputs)
         (
             x_options,
             y_multi_options,
@@ -286,8 +310,12 @@ def init_callbacks(dash_app):
         cs_style = {} if show_cs else {"display": "none"}
         show_treemap = chart_type == "treemap"
         treemap_style = {} if show_treemap else {"display": "none"}
+        show_funnel = chart_type == "funnel"
+        funnel_style = {} if show_funnel else {"display": "none"}
         standard_style = (
-            {"display": "none"} if show_map or show_cs or show_treemap else {}
+            {"display": "none"}
+            if show_map or show_cs or show_treemap or show_funnel
+            else {}
         )
         cscale_style = colorscale_input_style(chart_type=chart_type)
         drilldown_toggle_style = show_style((agg or "raw") != "raw")
@@ -304,9 +332,14 @@ def init_callbacks(dash_app):
             map_style,
             cs_style,
             treemap_style,
+            funnel_style,
             cscale_style,
             drilldown_toggle_style,
             lock_zoom_style(chart_type),
+            show_style(chart_type not in NON_EXT_AGGREGATION and len(y_val)),
+            "({} Selected)".format(len(inputs["cleaners"]))
+            if len(inputs["cleaners"])
+            else "",
         )
 
     @dash_app.callback(
@@ -517,7 +550,37 @@ def init_callbacks(dash_app):
             low_options,
         )
 
-    @dash_app.callback(
+    def label_value_callback(prop):
+        def _callback(selected_value, selected_label, group, data_id, **kwargs):
+            label_value_data = {
+                "{}_value".format(prop): selected_value,
+                "{}_label".format(prop): selected_label,
+            }
+            if group is not None:
+                label_value_data["{}_group".format(prop)] = group
+            label_value_data = dict_merge(label_value_data, kwargs)
+            df = global_state.get_data(data_id)
+            value_options, label_options = build_label_value_options(
+                df,
+                selected_value=selected_value,
+                selected_label=selected_label,
+            )
+            return label_value_data, value_options, label_options
+
+        return _callback
+
+    def funnel_callback(selected_value, selected_label, group, stacked, data_id):
+        label_value_data, value_options, label_options = label_value_callback("funnel")(
+            selected_value, selected_label, group, data_id, funnel_stacked=stacked
+        )
+        return (
+            label_value_data,
+            value_options,
+            label_options,
+            show_style(len(make_list(group)) > 0),
+        )
+
+    dash_app.callback(
         [
             Output("treemap-input-data", "data"),
             Output("treemap-value-dropdown", "options"),
@@ -529,26 +592,23 @@ def init_callbacks(dash_app):
             Input("treemap-group-dropdown", "value"),
         ],
         [State("data-tabs", "value")],
-    )
-    def treemap_data_callback(
-        treemap_value,
-        treemap_label,
-        group,
-        data_id,
-    ):
-        treemap_data = dict(
-            treemap_value=treemap_value,
-            treemap_label=treemap_label,
-        )
-        if group is not None:
-            treemap_data["treemap_group"] = group
-        df = global_state.get_data(data_id)
-        value_options, label_options = build_treemap_options(
-            df,
-            treemap_value=treemap_value,
-            treemap_label=treemap_label,
-        )
-        return treemap_data, value_options, label_options
+    )(label_value_callback("treemap"))
+
+    dash_app.callback(
+        [
+            Output("funnel-input-data", "data"),
+            Output("funnel-value-dropdown", "options"),
+            Output("funnel-label-dropdown", "options"),
+            Output("funnel-stack-input", "style"),
+        ],
+        [
+            Input("funnel-value-dropdown", "value"),
+            Input("funnel-label-dropdown", "value"),
+            Input("funnel-group-dropdown", "value"),
+            Input("funnel-stack-toggle", "on"),
+        ],
+        [State("data-tabs", "value")],
+    )(funnel_callback)
 
     @dash_app.callback(
         [
@@ -675,7 +735,21 @@ def init_callbacks(dash_app):
         final_is_open = is_open
         if n:
             final_is_open = not is_open
-        return final_is_open, data_selection_text(final_is_open)
+        return final_is_open, collapse_btn_text(final_is_open, text("Data Selection"))
+
+    @dash_app.callback(
+        [
+            Output("collapse-cleaners", "is_open"),
+            Output("collapse-cleaners-btn", "children"),
+        ],
+        [Input("collapse-cleaners-btn", "n_clicks")],
+        [State("collapse-cleaners", "is_open")],
+    )
+    def collapse_cleaners_input(n, is_open):
+        final_is_open = is_open
+        if n:
+            final_is_open = not is_open
+        return final_is_open, collapse_btn_text(final_is_open, text("Cleaners"))
 
     @dash_app.callback(
         [
@@ -686,6 +760,9 @@ def init_callbacks(dash_app):
             Output("yaxis-type", "children"),
             Output("load-clicks", "data"),
             Output("save-btn", "style"),
+            Output("agg-dropdown", "disabled"),
+            Output("extended-aggregation-tooltip", "children"),
+            Output("ext-agg-warning", "style"),
         ],
         # Since we use the data prop in an output,
         # we cannot get the initial data on load with the data prop.
@@ -700,6 +777,8 @@ def init_callbacks(dash_app):
             Input("map-input-data", "modified_timestamp"),
             Input("candlestick-input-data", "modified_timestamp"),
             Input("treemap-input-data", "modified_timestamp"),
+            Input("funnel-input-data", "modified_timestamp"),
+            Input("extended-aggregations", "modified_timestamp"),
             Input("load-btn", "n_clicks"),
         ],
         [
@@ -709,9 +788,11 @@ def init_callbacks(dash_app):
             State("map-input-data", "data"),
             State("candlestick-input-data", "data"),
             State("treemap-input-data", "data"),
+            State("funnel-input-data", "data"),
             State("last-chart-input-data", "data"),
             State("auto-load-toggle", "on"),
             State("load-clicks", "data"),
+            State("extended-aggregations", "data"),
         ],
     )
     def on_data(
@@ -721,16 +802,20 @@ def init_callbacks(dash_app):
         _ts4,
         _ts5,
         _ts6,
-        load,
+        _ts7,
+        _ts8,
+        load_clicks,
         inputs,
         chart_inputs,
         yaxis_data,
         map_data,
         cs_data,
         treemap_data,
+        funnel_data,
         last_chart_inputs,
         auto_load,
         prev_load_clicks,
+        ext_aggs,
     ):
         """
         dash callback controlling the building of dash charts
@@ -742,22 +827,54 @@ def init_callbacks(dash_app):
             map_data,
             cs_data,
             treemap_data,
+            funnel_data,
+            dict(extended_aggregation=ext_aggs or [])
+            if inputs.get("chart_type") not in NON_EXT_AGGREGATION
+            else {},
         )
-        if not auto_load and load == prev_load_clicks:
+        if not auto_load and load_clicks == prev_load_clicks:
             raise PreventUpdate
         if all_inputs == last_chart_inputs:
             raise PreventUpdate
         if is_app_root_defined(dash_app.server.config.get("APPLICATION_ROOT")):
             all_inputs["app_root"] = dash_app.server.config["APPLICATION_ROOT"]
         charts, range_data, code = build_chart(**all_inputs)
+        agg_disabled = len(ext_aggs) > 0
+        ext_agg_tt = text("ext_agg_desc")
+        ext_agg_warning = show_style(agg_disabled)
+        if agg_disabled:
+            ext_agg_tt = html.Div(
+                [
+                    html.Span(text("ext_agg_desc")),
+                    html.Br(),
+                    html.Ul(
+                        [
+                            html.Li(
+                                extended_aggregations.build_extended_agg_desc(ext_agg),
+                                className="mb-0",
+                            )
+                            for ext_agg in ext_aggs
+                        ]
+                    ),
+                ]
+            )
+        final_cols = build_final_cols(
+            make_list(inputs.get("y")),
+            inputs.get("z"),
+            inputs.get("agg"),
+            ext_aggs if inputs.get("chart_type") not in NON_EXT_AGGREGATION else [],
+        )
         return (
             charts,
             all_inputs,
             range_data,
             "\n".join(make_list(code) + [CHART_EXPORT_CODE]),
-            get_yaxis_type_tabs(make_list(inputs.get("y") or [])),
-            load,
+            get_yaxis_type_tabs(final_cols),
+            load_clicks,
             dict(display="block" if valid_chart(**all_inputs) else "none"),
+            agg_disabled,
+            ext_agg_tt,
+            ext_agg_warning,
         )
 
     def get_default_range(range_data, y, max=False):
@@ -790,15 +907,24 @@ def init_callbacks(dash_app):
             State("input-data", "data"),
             State("yaxis-data", "data"),
             State("range-data", "data"),
+            State("extended-aggregations", "data"),
         ],
     )
-    def yaxis_min_max_values(yaxis_type, yaxis, inputs, yaxis_inputs, range_data):
+    def yaxis_min_max_values(
+        yaxis_type, yaxis, inputs, yaxis_inputs, range_data, ext_aggs
+    ):
         """
         dash callback controlling values for selected y-axis in y-axis range editor
         """
         y = make_list(inputs.get("y"))
+        final_cols = build_final_cols(
+            y,
+            inputs.get("z"),
+            inputs.get("agg"),
+            ext_aggs if inputs.get("chart_type") not in NON_EXT_AGGREGATION else [],
+        )
         dd_style = dict(
-            display="block" if yaxis_type == "multi" and len(y) > 1 else "none"
+            display="block" if yaxis_type == "multi" and len(final_cols) > 1 else "none"
         )
         type_style = (
             {"borderRadius": "0 0.25rem 0.25rem 0"} if yaxis_type == "default" else None
@@ -820,10 +946,10 @@ def init_callbacks(dash_app):
             curr_vals = (yaxis_inputs or {}).get("data", {}).get("all") or {}
             curr_min = curr_vals.get("min")
             if curr_min is None:
-                curr_min = get_default_range(range_min, y)
+                curr_min = get_default_range(range_min, final_cols)
             curr_max = curr_vals.get("max")
             if curr_max is None:
-                curr_max = get_default_range(range_max, y, max=True)
+                curr_max = get_default_range(range_max, final_cols, max=True)
         return (
             curr_min,
             curr_max,
@@ -844,26 +970,42 @@ def init_callbacks(dash_app):
         ],
         [
             State("yaxis-dropdown", "value"),
+            State("yaxis-dropdown", "options"),
             State("yaxis-data", "data"),
             State("range-data", "data"),
             State("input-data", "data"),
+            State("extended-aggregations", "data"),
         ],
     )
     def update_yaxis_data(
-        yaxis_type, yaxis_min, yaxis_max, yaxis, yaxis_data, range_data, inputs
+        yaxis_type,
+        yaxis_min,
+        yaxis_max,
+        yaxis,
+        yaxes,
+        yaxis_data,
+        range_data,
+        inputs,
+        ext_aggs,
     ):
         """
         dash callback controlling updates to y-axis range state
         """
         yaxis_data = yaxis_data or dict(data={})
         yaxis_data["type"] = yaxis_type
+        yaxis = yaxis or yaxes[0]["value"] if len(yaxes) else None
         yaxis_name = "all" if yaxis_type == "single" else yaxis
         if yaxis_name == "all":
-            y = make_list(inputs.get("y"))
+            final_cols = build_final_cols(
+                make_list(inputs.get("y")),
+                inputs.get("z"),
+                inputs.get("agg"),
+                ext_aggs if inputs.get("chart_type") not in NON_EXT_AGGREGATION else [],
+            )
             mins = range_data.get("min", {})
             maxs = range_data.get("max", {})
-            range_min = get_default_range(mins, y)
-            range_max = get_default_range(maxs, y, max=True)
+            range_min = get_default_range(mins, final_cols)
+            range_max = get_default_range(maxs, final_cols, max=True)
         elif yaxis is None:
             raise PreventUpdate
         else:
@@ -894,19 +1036,30 @@ def init_callbacks(dash_app):
             Input("map-input-data", "modified_timestamp"),
             Input("candlestick-input-data", "modified_timestamp"),
             Input("treemap-input-data", "modified_timestamp"),
+            Input("funnel-input-data", "modified_timestamp"),
         ],
         [
             State("input-data", "data"),
             State("map-input-data", "data"),
             State("candlestick-input-data", "data"),
             State("treemap-input-data", "data"),
+            State("funnel-input-data", "data"),
         ],
     )
     def main_input_class(
-        _ts, _ts2, _ts3, _ts4, inputs, map_inputs, cs_inputs, treemap_inputs
+        _ts,
+        _ts2,
+        _ts3,
+        _ts4,
+        _ts5,
+        inputs,
+        map_inputs,
+        cs_inputs,
+        treemap_inputs,
+        funnel_inputs,
     ):
         return main_inputs_and_group_val_display(
-            dict_merge(inputs, map_inputs, cs_inputs, treemap_inputs)
+            dict_merge(inputs, map_inputs, cs_inputs, treemap_inputs, funnel_inputs)
         )
 
     @dash_app.callback(
@@ -920,6 +1073,7 @@ def init_callbacks(dash_app):
             Input("map-group-dropdown", "value"),
             Input("candlestick-group-dropdown", "value"),
             Input("treemap-group-dropdown", "value"),
+            Input("funnel-group-dropdown", "value"),
         ],
         [
             State("input-data", "data"),
@@ -932,6 +1086,7 @@ def init_callbacks(dash_app):
         map_group_cols,
         cs_group_cols,
         treemap_group_cols,
+        funnel_group_cols,
         inputs,
         prev_group_vals,
     ):
@@ -943,6 +1098,8 @@ def init_callbacks(dash_app):
             group_cols = cs_group_cols
         elif chart_type == "treemap":
             group_cols = treemap_group_cols
+        elif chart_type == "funnel":
+            group_cols = funnel_group_cols
         group_cols = make_list(group_cols)
         group_types = get_group_types(inputs, group_cols)
         if "groups" not in group_types:
@@ -983,5 +1140,6 @@ def init_callbacks(dash_app):
 
     custom_geojson.init_callbacks(dash_app)
     drilldown_modal.init_callbacks(dash_app)
+    extended_aggregations.init_callbacks(dash_app)
     lock_zoom.init_callbacks(dash_app)
     saved_charts.init_callbacks(dash_app)
